<?php

namespace Drupal\views_grouping_field_pager\Plugin\views\pager;

use Drupal\Core\Form\FormStateInterface;
use Drupal\views\Plugin\views\pager\Full;
use Drupal\views\ViewExecutable;
use Drupal\views\Views;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\views\Attribute\ViewsPager;

/**
 * Pager for views grouped by a field.
 *
 * @ingroup views_pager_plugins
 */
#[ViewsPager(
  id: "grouping_field_pager",
  title: new TranslatableMarkup("Paged output, full pager, for manipulating a view grouped by a field"),
  short_title: new TranslatableMarkup("Grouping field pager"),
  help: new TranslatableMarkup("Paged output, full pager, for manipulating a view grouped by a field"),
  theme: "pager",
  register_theme: FALSE
)]
class GroupingFieldPager extends Full
{
  protected $pages = [];

  protected $groupCount = 0;

  /**
   * {@inheritdoc}
   */
  protected function defineOptions() {
    $options = parent::defineOptions();

    $options['grouping_field'] = [
      'default' => '',
    ];

    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function buildOptionsForm(&$form, FormStateInterface $form_state) {
    parent::buildOptionsForm($form, $form_state);

    // Add a field to configure the grouping field.
    $form['grouping_field'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Grouping field'),
      '#description' => $this->t('Enter the machine name of the field to group results by.'),
      '#default_value' => $this->options['grouping_field'],
    ];
  }


  /**
   * {@inheritdoc}
   */
  public function query()
  {
    parent::query();

    if (($items_per_page = (int) $this->getItemsPerPage()) < 1) {
      return;
    }

    $count_view = $this->executeCountView();
    if (empty($count_view?->result)) {
      return;
    }

    if (empty($grouping_field = $this->getGroupingFieldFromView($count_view))) {
      return;
    }

    // Break up count view results into groups.
    foreach ($count_view->result as $i => $row) {
      $groups[$row->$grouping_field][$i] = $row;
    }

    // Break up count view groups into pages.
    $this->pages = array_chunk($groups, $items_per_page, TRUE);

    $offset = 0;
    $limit = 0;
    $this->groupCount = count($groups);

    // Calculate offset and limit for the actual view.
    foreach ($this->pages as $page_num => $groups) {
      $on_current_page = FALSE;

      foreach ($groups as $results) {
        if ($page_num === $this->getCurrentPage()) {
          $on_current_page = TRUE;
          break;
        }
        $offset += count($results);
      }

      if ($on_current_page) {
        foreach ($groups as $results) {
          $limit += count($results);
        }
        break;
      }
    }

    // Alter view query.
    $this->view->query->setOffset($offset);
    $this->view->query->setLimit($limit);
  }

  public function getTotalItems() {
    return $this->groupCount;
  }

  protected function executeCountView(): ?ViewExecutable {
    if ($count_view = Views::getView($this->view->id())) {
      $count_view->setDisplay($this->view->current_display);
      $count_view->display_handler->overrideOption('pager', [
        'type' => 'none',
        'options' => [
          'items_per_page' => 0,
          'offset' => 0,
        ],
      ]);
      $count_view->setArguments($this->view->args);
      $count_view->preExecute();
      $count_view->execute();

      return $count_view;
    }

    return NULL;
  }

  protected function getGroupingFieldFromView(ViewExecutable $view): ?string {
    if (!empty($this->options['grouping_field'])) {
      return $this->options['grouping_field'];
    }

    // Guess grouping field if it is not set.
    $display_options = $view->display_handler->getPlugin('style')->options;
    if (!empty($display_options['grouping'])
      && is_array($display_options['grouping'])) {
      $grouping_config = reset($display_options['grouping']);
      if (!empty($grouping_config['field'])) {
        // @fixme There is surely a better way of getting the fully qualified name.
        foreach (get_object_vars($view->result[0]) as $property_name => $value) {
          if (str_contains($property_name, '__' . $grouping_config['field'] . '_')
            && !str_ends_with($property_name, '_langcode')) {
            return $property_name;
          }
        }
      }
    }

    return NULL;
  }
}
